structure BinaryTree_Code : sig
  val id : 'a -> 'a
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype num = One | Bit0 of num | Bit1 of num
  datatype inta = Zero_int | Pos of num | Neg of num
  datatype 'a tree = Tip | T of 'a tree * 'a * 'a tree
  val equal_num : num -> num -> bool
  val equal_inta : inta -> inta -> bool
  val equal_int : inta equal
  val less_num : num -> num -> bool
  val less_eq_num : num -> num -> bool
  val less_int : inta -> inta -> bool
  val fst : 'a * 'b -> 'a
  val snd : 'a * 'b -> 'b
  val equal_Tree : 'a equal -> 'a tree -> 'a tree -> bool
  val wrmrm : 'a equal -> ('a -> inta) -> 'a tree -> 'a tree * 'a
  val remove : 'a equal -> ('a -> inta) -> 'a -> 'a tree -> 'a tree
  val binsert : ('a -> inta) -> 'a -> 'a tree -> 'a tree
  val tlookup : ('a -> inta) -> inta -> 'a tree -> 'a option
  val test : inta option
end = struct

fun id x = (fn xa => xa) x;

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype num = One | Bit0 of num | Bit1 of num;

datatype inta = Zero_int | Pos of num | Neg of num;

datatype 'a tree = Tip | T of 'a tree * 'a * 'a tree;

fun equal_num (Bit1 numa) (Bit0 num) = false
  | equal_num (Bit0 numa) (Bit1 num) = false
  | equal_num (Bit1 num) One = false
  | equal_num One (Bit1 num) = false
  | equal_num (Bit0 num) One = false
  | equal_num One (Bit0 num) = false
  | equal_num (Bit1 numa) (Bit1 num) = equal_num numa num
  | equal_num (Bit0 numa) (Bit0 num) = equal_num numa num
  | equal_num One One = true;

fun equal_inta (Neg k) (Neg l) = equal_num k l
  | equal_inta (Neg k) (Pos l) = false
  | equal_inta (Neg k) Zero_int = false
  | equal_inta (Pos k) (Neg l) = false
  | equal_inta (Pos k) (Pos l) = equal_num k l
  | equal_inta (Pos k) Zero_int = false
  | equal_inta Zero_int (Neg l) = false
  | equal_inta Zero_int (Pos l) = false
  | equal_inta Zero_int Zero_int = true;

val equal_int = {equal = equal_inta} : inta equal;

fun less_num (Bit1 m) (Bit0 n) = less_num m n
  | less_num (Bit1 m) (Bit1 n) = less_num m n
  | less_num (Bit0 m) (Bit1 n) = less_eq_num m n
  | less_num (Bit0 m) (Bit0 n) = less_num m n
  | less_num One (Bit1 n) = true
  | less_num One (Bit0 n) = true
  | less_num m One = false
and less_eq_num (Bit1 m) (Bit0 n) = less_num m n
  | less_eq_num (Bit1 m) (Bit1 n) = less_eq_num m n
  | less_eq_num (Bit0 m) (Bit1 n) = less_eq_num m n
  | less_eq_num (Bit0 m) (Bit0 n) = less_eq_num m n
  | less_eq_num (Bit1 m) One = false
  | less_eq_num (Bit0 m) One = false
  | less_eq_num One n = true;

fun less_int (Neg k) (Neg l) = less_num l k
  | less_int (Neg k) (Pos l) = true
  | less_int (Neg k) Zero_int = true
  | less_int (Pos k) (Neg l) = false
  | less_int (Pos k) (Pos l) = less_num k l
  | less_int (Pos k) Zero_int = false
  | less_int Zero_int (Neg l) = false
  | less_int Zero_int (Pos l) = true
  | less_int Zero_int Zero_int = false;

fun fst (a, b) = a;

fun snd (a, b) = b;

fun equal_Tree A_ (T (tree1, a, tree2)) Tip = false
  | equal_Tree A_ Tip (T (tree1, a, tree2)) = false
  | equal_Tree A_ (T (tree1a, aa, tree2a)) (T (tree1, a, tree2)) =
    equal_Tree A_ tree1a tree1 andalso
      (eq A_ aa a andalso equal_Tree A_ tree2a tree2)
  | equal_Tree A_ Tip Tip = true;

fun wrmrm A_ h (T (t1, x, t2)) =
  (if equal_Tree A_ t2 Tip then (t1, x)
    else (T (t1, x, fst (wrmrm A_ h t2)), snd (wrmrm A_ h t2)));

fun remove A_ h e Tip = Tip
  | remove A_ h e (T (t1, x, t2)) =
    (if less_int (h e) (h x) then T (remove A_ h e t1, x, t2)
      else (if less_int (h x) (h e) then T (t1, x, remove A_ h e t2)
             else (if equal_Tree A_ t1 Tip then t2
                    else let
                           val (t1p, r) = wrmrm A_ h t1;
                         in
                           T (t1p, r, t2)
                         end)));

fun binsert h e Tip = T (Tip, e, Tip)
  | binsert h e (T (t1, x, t2)) =
    (if less_int (h e) (h x) then T (binsert h e t1, x, t2)
      else (if less_int (h x) (h e) then T (t1, x, binsert h e t2)
             else T (t1, e, t2)));

fun tlookup h k Tip = NONE
  | tlookup h k (T (t1, x, t2)) =
    (if less_int k (h x) then tlookup h k t1
      else (if less_int (h x) k then tlookup h k t2 else SOME x));

val test : inta option =
  tlookup id (Pos (Bit0 (Bit0 One)))
    (remove equal_int id (Pos (Bit1 One))
      (binsert id (Pos (Bit0 (Bit0 One))) (binsert id (Pos (Bit1 One)) Tip)));

end; (*struct BinaryTree_Code*)
